@@ -10,6 +10,7 @@
 #include "config.h"
 #include "log.h"
 #include "nelem.h"
+#include "scale.h"
 #include "unicode.h"
 #include "xmalloc.h"
 
@@ -74,10 +75,10 @@ struct uint32_percent {
 static char *strip(const char *str);
 static bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const char *option, const char *value);
 static char *get_config_path(void);
-static uint32_t fixup_percentage(uint32_t value, uint32_t base, bool is_percent, uint32_t scale, bool use_scale);
-static void fixup_text_theme(struct text_theme *theme, uint32_t scale);
+static uint32_t fixup_percentage(uint32_t value, uint32_t base, bool is_percent);
 
 static uint32_t parse_anchor(const char *filename, size_t lineno, const char *str, bool *err);
+static enum cursor_style parse_cursor_style(const char *filename, size_t lineno, const char *str, bool *err);
 static bool parse_bool(const char *filename, size_t lineno, const char *str, bool *err);
 static uint32_t parse_char(const char *filename, size_t lineno, const char *str, bool *err);
 static struct color parse_color(const char *filename, size_t lineno, const char *str, bool *err);
@@ -334,7 +335,11 @@ bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const
 	} else if (strcasecmp(option, "output") == 0) {
 		snprintf(tofi->target_output_name, N_ELEM(tofi->target_output_name), "%s", value);
 	} else if (strcasecmp(option, "font") == 0) {
-		snprintf(tofi->window.entry.font_name, N_ELEM(tofi->window.entry.font_name), "%s", value);
+		if ((strlen(value) > 2) && (value[0] == '~') && (value[1] == '/')) {
+			snprintf(tofi->window.entry.font_name, N_ELEM(tofi->window.entry.font_name), "%s%s", getenv("HOME"), &(value[1]));
+		} else {
+			snprintf(tofi->window.entry.font_name, N_ELEM(tofi->window.entry.font_name), "%s", value);
+		}
 	} else if (strcasecmp(option, "font-size") == 0) {
 		uint32_t val =  parse_uint32(filename, lineno, value, &err);
 		if (val == 0) {
@@ -362,6 +367,39 @@ bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const
 		if (!err) {
 			tofi->window.entry.outline_color = val;
 		}
+	} else if (strcasecmp(option, "text-cursor") == 0) {
+		bool val = parse_bool(filename, lineno, value, &err);
+		if (!err) {
+			tofi->window.entry.cursor_theme.show = val;
+		}
+	} else if (strcasecmp(option, "text-cursor-style") == 0) {
+		enum cursor_style val = parse_cursor_style(filename, lineno, value, &err);
+		if (!err) {
+			tofi->window.entry.cursor_theme.style = val;
+		}
+	} else if (strcasecmp(option, "text-cursor-color") == 0) {
+		struct color val = parse_color(filename, lineno, value, &err);
+		if (!err) {
+			tofi->window.entry.cursor_theme.color = val;
+			tofi->window.entry.cursor_theme.color_specified = true;
+		}
+	} else if (strcasecmp(option, "text-cursor-background") == 0) {
+		struct color val = parse_color(filename, lineno, value, &err);
+		if (!err) {
+			tofi->window.entry.cursor_theme.text_color = val;
+			tofi->window.entry.cursor_theme.text_color_specified = true;
+		}
+	} else if (strcasecmp(option, "text-cursor-corner-radius") == 0) {
+		uint32_t val = parse_uint32(filename, lineno, value, &err);
+		if (!err) {
+			tofi->window.entry.cursor_theme.corner_radius = val;
+		}
+	} else if (strcasecmp(option, "text-cursor-thickness") == 0) {
+		uint32_t val = parse_uint32(filename, lineno, value, &err);
+		if (!err) {
+			tofi->window.entry.cursor_theme.thickness = val;
+			tofi->window.entry.cursor_theme.thickness_specified = true;
+		}
 	} else if (strcasecmp(option, "prompt-text") == 0) {
 		snprintf(tofi->window.entry.prompt_text, N_ELEM(tofi->window.entry.prompt_text), "%s", value);
 	} else if (strcasecmp(option, "prompt-padding") == 0) {
@@ -655,6 +693,11 @@ bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const
 		if (!err) {
 			tofi->require_match = val;
 		}
+	} else if (strcasecmp(option, "auto-accept-single") == 0) {
+		bool val = parse_bool(filename, lineno, value, &err);
+		if (!err) {
+			tofi->auto_accept_single = val;
+		}
 	} else if (strcasecmp(option, "hide-input") == 0) {
 		bool val = parse_bool(filename, lineno, value, &err);
 		if (!err) {
@@ -710,116 +753,92 @@ bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const
 	return !err;
 }
 
-void config_apply(struct tofi *tofi, const char *option, const char *value)
+bool config_apply(struct tofi *tofi, const char *option, const char *value)
 {
-	if (!parse_option(tofi, "", 0, option, value)) {
-		exit(EXIT_FAILURE);
-	};
+	return parse_option(tofi, "", 0, option, value);
 }
 
-uint32_t fixup_percentage(uint32_t value, uint32_t base, bool is_percent, uint32_t scale, bool use_scale)
+uint32_t fixup_percentage(uint32_t value, uint32_t base, bool is_percent)
 {
 	if (is_percent) {
 		return value * base / 100;
-	} else if (use_scale) {
-		return value * scale;
 	}
 	return value;
 }
 
-void fixup_text_theme(struct text_theme *theme, uint32_t scale)
-{
-	theme->background_corner_radius *= scale;
-	theme->padding.top *= scale;
-	theme->padding.bottom *= scale;
-	theme->padding.left *= scale;
-	theme->padding.right *= scale;
-}
-
 void config_fixup_values(struct tofi *tofi)
 {
-	uint32_t scale = tofi->window.scale;
-
-	if (tofi->use_scale) {
-		struct entry *entry = &tofi->window.entry;
-
-		entry->font_size *= scale;
-		entry->prompt_padding *= scale;
-		entry->corner_radius *= scale;
-		entry->result_spacing *= scale;
-		entry->input_width *= scale;
-		entry->outline_width *= scale;
-		entry->border_width *= scale;
-
-		fixup_text_theme(&entry->prompt_theme, scale);
-		fixup_text_theme(&entry->placeholder_theme, scale);
-		fixup_text_theme(&entry->input_theme, scale);
-		fixup_text_theme(&entry->default_result_theme, scale);
-		fixup_text_theme(&entry->alternate_result_theme, scale);
-		fixup_text_theme(&entry->selection_theme, scale);
+	uint32_t base_width = tofi->output_width;
+	uint32_t base_height = tofi->output_height;
+	uint32_t scale;
+	if (tofi->window.fractional_scale != 0) {
+		scale = tofi->window.fractional_scale;
+	} else {
+		scale = tofi->window.scale * 120;
 	}
 
-	/* These values should only be scaled if they're not percentages. */
+	/*
+	 * If we're going to be scaling these values in Cairo,
+	 * we need to apply the inverse scale here.
+	 */
+	if (tofi->use_scale) {
+		base_width = scale_apply_inverse(base_width, scale);
+		base_height = scale_apply_inverse(base_height, scale);
+	}
+
+	tofi->window.margin_top = fixup_percentage(
+			tofi->window.margin_top,
+			base_height,
+			tofi->window.margin_top_is_percent);
+	tofi->window.margin_bottom = fixup_percentage(
+			tofi->window.margin_bottom,
+			base_height,
+			tofi->window.margin_bottom_is_percent);
+	tofi->window.margin_left = fixup_percentage(
+			tofi->window.margin_left,
+			base_width,
+			tofi->window.margin_left_is_percent);
+	tofi->window.margin_right = fixup_percentage(
+			tofi->window.margin_right,
+			base_width,
+			tofi->window.margin_right_is_percent);
+	tofi->window.entry.padding_top = fixup_percentage(
+			tofi->window.entry.padding_top,
+			base_height,
+			tofi->window.entry.padding_top_is_percent);
+	tofi->window.entry.padding_bottom = fixup_percentage(
+			tofi->window.entry.padding_bottom,
+			base_height,
+			tofi->window.entry.padding_bottom_is_percent);
+	tofi->window.entry.padding_left = fixup_percentage(
+			tofi->window.entry.padding_left,
+			base_width,
+			tofi->window.entry.padding_left_is_percent);
+	tofi->window.entry.padding_right = fixup_percentage(
+			tofi->window.entry.padding_right,
+			base_width,
+			tofi->window.entry.padding_right_is_percent);
+
+	/*
+	 * Window width and height are a little special. We're only going to be
+	 * using them to specify sizes to Wayland, which always wants scaled
+	 * pixels, so always scale them here (unless we've directly specified a
+	 * scaled size).
+	 */
 	tofi->window.width = fixup_percentage(
 			tofi->window.width,
 			tofi->output_width,
-			tofi->window.width_is_percent,
-			tofi->window.scale,
-			tofi->use_scale);
+			tofi->window.width_is_percent);
 	tofi->window.height = fixup_percentage(
 			tofi->window.height,
 			tofi->output_height,
-			tofi->window.height_is_percent,
-			tofi->window.scale,
-			tofi->use_scale);
-	tofi->window.margin_top = fixup_percentage(
-			tofi->window.margin_top,
-			tofi->output_height,
-			tofi->window.margin_top_is_percent,
-			tofi->window.scale,
-			tofi->use_scale);
-	tofi->window.margin_bottom = fixup_percentage(
-			tofi->window.margin_bottom,
-			tofi->output_height,
-			tofi->window.margin_bottom_is_percent,
-			tofi->window.scale,
-			tofi->use_scale);
-	tofi->window.margin_left = fixup_percentage(
-			tofi->window.margin_left,
-			tofi->output_width,
-			tofi->window.margin_left_is_percent,
-			tofi->window.scale,
-			tofi->use_scale);
-	tofi->window.margin_right = fixup_percentage(
-			tofi->window.margin_right,
-			tofi->output_width,
-			tofi->window.margin_right_is_percent,
-			tofi->window.scale,
-			tofi->use_scale);
-	tofi->window.entry.padding_top = fixup_percentage(
-			tofi->window.entry.padding_top,
-			tofi->output_height,
-			tofi->window.entry.padding_top_is_percent,
-			tofi->window.scale,
-			tofi->use_scale);
-	tofi->window.entry.padding_bottom = fixup_percentage(
-			tofi->window.entry.padding_bottom,
-			tofi->output_height,
-			tofi->window.entry.padding_bottom_is_percent,
-			tofi->window.scale,
-			tofi->use_scale);
-	tofi->window.entry.padding_left = fixup_percentage(
-			tofi->window.entry.padding_left,
-			tofi->output_width,
-			tofi->window.entry.padding_left_is_percent,
-			tofi->window.scale,
-			tofi->use_scale);
-	tofi->window.entry.padding_right = fixup_percentage(
-			tofi->window.entry.padding_right,
-			tofi->output_width,
-			tofi->window.entry.padding_right_is_percent,
-			tofi->window.scale,
-			tofi->use_scale);
+			tofi->window.height_is_percent);
+	if (tofi->window.width_is_percent || !tofi->use_scale) {
+		tofi->window.width = scale_apply_inverse(tofi->window.width, scale);
+	}
+	if (tofi->window.height_is_percent || !tofi->use_scale) {
+		tofi->window.height = scale_apply_inverse(tofi->window.height, scale);
+	}
 
 	/* Don't attempt percentage handling if exclusive_zone is set to -1. */
 	if (tofi->window.exclusive_zone > 0) {
@@ -829,24 +848,20 @@ void config_fixup_values(struct tofi *tofi)
 			case ANCHOR_BOTTOM:
 				tofi->window.exclusive_zone = fixup_percentage(
 						tofi->window.exclusive_zone,
-						tofi->output_height,
-						tofi->window.exclusive_zone_is_percent,
-						tofi->window.scale,
-						tofi->use_scale);
+						base_height,
+						tofi->window.exclusive_zone_is_percent);
 				break;
 			case ANCHOR_LEFT:
 			case ANCHOR_RIGHT:
 				tofi->window.exclusive_zone = fixup_percentage(
 						tofi->window.exclusive_zone,
-						tofi->output_width,
-						tofi->window.exclusive_zone_is_percent,
-						tofi->window.scale,
-						tofi->use_scale);
+						base_width,
+						tofi->window.exclusive_zone_is_percent);
 				break;
 			default:
 				/*
-				 * Exclusive zone >0 is meaningless for other anchor
-				 * positions.
+				 * Exclusive zone >0 is meaningless for other
+				 * anchor positions.
 				 */
 				tofi->window.exclusive_zone =
 					MIN(tofi->window.exclusive_zone, 0);
@@ -905,6 +920,9 @@ uint32_t parse_char(const char *filename, size_t lineno, const char *str, bool *
 	ch = utf8_to_utf32_validate(tmp);
 	if (ch == (uint32_t)-2 || ch == (uint32_t)-1 || *utf8_next_char(tmp) != '\0') {
 		PARSE_ERROR(filename, lineno, "Failed to parse \"%s\" as a character.\n", str);
+		if (err) {
+			*err = true;
+		}
 	}
 	free(tmp);
 
@@ -941,7 +959,27 @@ uint32_t parse_anchor(const char *filename, size_t lineno, const char *str, bool
 		return ANCHOR_CENTER;
 	}
 	PARSE_ERROR(filename, lineno, "Invalid anchor \"%s\".\n", str);
-	*err = true;
+	if (err) {
+		*err = true;
+	}
+	return 0;
+}
+
+enum cursor_style parse_cursor_style(const char *filename, size_t lineno, const char *str, bool *err)
+{
+	if(strcasecmp(str, "bar") == 0) {
+		return CURSOR_STYLE_BAR;
+	}
+	if(strcasecmp(str, "block") == 0) {
+		return CURSOR_STYLE_BLOCK;
+	}
+	if(strcasecmp(str, "underscore") == 0) {
+		return CURSOR_STYLE_UNDERSCORE;
+	}
+	PARSE_ERROR(filename, lineno, "Invalid cursor style \"%s\".\n", str);
+	if (err) {
+		*err = true;
+	}
 	return 0;
 }
 
@@ -961,13 +999,13 @@ uint32_t parse_uint32(const char *filename, size_t lineno, const char *str, bool
 {
 	errno = 0;
 	char *endptr;
-	int32_t ret = strtoul(str, &endptr, 0);
-	if (endptr == str) {
+	int64_t ret = strtoul(str, &endptr, 0);
+	if (endptr == str || *endptr != '\0') {
 		PARSE_ERROR(filename, lineno, "Failed to parse \"%s\" as unsigned int.\n", str);
 		if (err) {
 			*err = true;
 		}
-	} else if (errno || ret < 0) {
+	} else if (errno || ret < 0 || ret > UINT32_MAX) {
 		PARSE_ERROR(filename, lineno, "Unsigned int value \"%s\" out of range.\n", str);
 		if (err) {
 			*err = true;
@@ -980,13 +1018,13 @@ int32_t parse_int32(const char *filename, size_t lineno, const char *str, bool *
 {
 	errno = 0;
 	char *endptr;
-	int32_t ret = strtol(str, &endptr, 0);
-	if (endptr == str) {
+	int64_t ret = strtol(str, &endptr, 0);
+	if (endptr == str || *endptr != '\0') {
 		PARSE_ERROR(filename, lineno, "Failed to parse \"%s\" as int.\n", str);
 		if (err) {
 			*err = true;
 		}
-	} else if (errno) {
+	} else if (errno || ret < INT32_MIN || ret > INT32_MAX) {
 		PARSE_ERROR(filename, lineno, "Int value \"%s\" out of range.\n", str);
 		if (err) {
 			*err = true;
@@ -999,14 +1037,14 @@ struct uint32_percent parse_uint32_percent(const char *filename, size_t lineno,
 {
 	errno = 0;
 	char *endptr;
-	int32_t val = strtoul(str, &endptr, 0);
+	int64_t val = strtoul(str, &endptr, 0);
 	bool percent = false;
-	if (endptr == str) {
+	if (endptr == str || (*endptr != '\0' && *endptr != '%')) {
 		PARSE_ERROR(filename, lineno, "Failed to parse \"%s\" as unsigned int.\n", str);
 		if (err) {
 			*err = true;
 		}
-	} else if (errno || val < 0) {
+	} else if (errno || val < 0 || val > UINT32_MAX) {
 		PARSE_ERROR(filename, lineno, "Unsigned int value \"%s\" out of range.\n", str);
 		if (err) {
 			*err = true;
@@ -1022,7 +1060,8 @@ struct uint32_percent parse_uint32_percent(const char *filename, size_t lineno,
 
 struct directional parse_directional(const char *filename, size_t lineno, const char *str, bool *err)
 {
-	int32_t values[4];
+	/* One extra value to act as a sentinel for too many values in str. */
+	int32_t values[5];
 	char *saveptr = NULL;
 	char *tmp = xstrdup(str);
 	char *val = strtok_r(tmp, ",", &saveptr);
