@@ -4,6 +4,7 @@
 #include "entry.h"
 #include "log.h"
 #include "nelem.h"
+#include "scale.h"
 
 #undef MAX
 #define MAX(a, b) ((a) > (b) ? (a) : (b))
@@ -43,27 +44,9 @@ static void apply_text_theme_fallback(struct text_theme *theme, const struct tex
 	}
 }
 
-static void fixup_padding_sizes(struct directional *padding, uint32_t clip_width, uint32_t clip_height)
+void entry_init(struct entry *entry, uint8_t *restrict buffer, uint32_t width, uint32_t height, uint32_t fractional_scale_numerator)
 {
-	if (padding->top < 0) {
-		padding->top = clip_height;
-	}
-	if (padding->bottom < 0) {
-		padding->bottom = clip_height;
-	}
-	if (padding->left < 0) {
-		padding->left = clip_width;
-	}
-	if (padding->right < 0) {
-		padding->right = clip_width;
-	}
-}
-
-void entry_init(struct entry *entry, uint8_t *restrict buffer, uint32_t width, uint32_t height)
-{
-	entry->image.width = width;
-	entry->image.height = height;
-
+	double scale = fractional_scale_numerator / 120.;
 	/*
 	 * Create the cairo surfaces and contexts we'll be using.
 	 *
@@ -73,6 +56,10 @@ void entry_init(struct entry *entry, uint8_t *restrict buffer, uint32_t width, u
 	 * (width * height * (sizeof(uint32_t) == 4) * 2) bytes,
 	 * to allow for double buffering.
 	 */
+	log_debug("Creating %u x %u Cairo surface with scale factor %.3lf.\n",
+			width,
+			height,
+			fractional_scale_numerator / 120.);
 	cairo_surface_t *surface = cairo_image_surface_create_for_data(
 			buffer,
 			CAIRO_FORMAT_ARGB32,
@@ -80,6 +67,7 @@ void entry_init(struct entry *entry, uint8_t *restrict buffer, uint32_t width, u
 			height,
 			width * sizeof(uint32_t)
 			);
+	cairo_surface_set_device_scale(surface, scale, scale);
 	cairo_t *cr = cairo_create(surface);
 
 	entry->cairo[0].surface = surface;
@@ -92,8 +80,12 @@ void entry_init(struct entry *entry, uint8_t *restrict buffer, uint32_t width, u
 			height,
 			width * sizeof(uint32_t)
 			);
+	cairo_surface_set_device_scale(entry->cairo[1].surface, scale, scale);
 	entry->cairo[1].cr = cairo_create(entry->cairo[1].surface);
 
+	/* If we're scaling with Cairo, remember to account for that here. */
+	width = scale_apply_inverse(width, fractional_scale_numerator);
+	height = scale_apply_inverse(height, fractional_scale_numerator);
 
 	log_debug("Drawing window.\n");
 	/* Draw the background */
@@ -121,8 +113,14 @@ void entry_init(struct entry *entry, uint8_t *restrict buffer, uint32_t width, u
 	cairo_stroke_preserve(cr);
 
 	/* Clear the overdrawn bits outside of the rounded corners */
-	cairo_rectangle(cr, 0, 0, width, height);
-	cairo_set_source_rgba(cr, 0, 0, 0, 0.5);
+	/*
+	 * N.B. the +1's shouldn't be required, but certain fractional scale
+	 * factors can otherwise cause 1-pixel artifacts on the edges
+	 * (presumably because Cairo is performing rounding differently to us
+	 * at some point).
+	 */
+	cairo_rectangle(cr, 0, 0, width + 1, height + 1);
+	cairo_set_source_rgba(cr, 0, 0, 0, 1);
 	cairo_save(cr);
 	cairo_set_fill_rule(cr, CAIRO_FILL_RULE_EVEN_ODD);
 	cairo_set_operator(cr, CAIRO_OPERATOR_CLEAR);
@@ -204,12 +202,13 @@ void entry_init(struct entry *entry, uint8_t *restrict buffer, uint32_t width, u
 	apply_text_theme_fallback(&entry->alternate_result_theme, &entry->default_result_theme);
 	apply_text_theme_fallback(&entry->selection_theme, &default_theme);
 
-	fixup_padding_sizes(&entry->prompt_theme.padding, width, height);
-	fixup_padding_sizes(&entry->input_theme.padding, width, height);
-	fixup_padding_sizes(&entry->placeholder_theme.padding, width, height);
-	fixup_padding_sizes(&entry->default_result_theme.padding, width, height);
-	fixup_padding_sizes(&entry->alternate_result_theme.padding, width, height);
-	fixup_padding_sizes(&entry->selection_theme.padding, width, height);
+	/* The cursor is a special case, as it just needs the input colours. */
+	if (!entry->cursor_theme.color_specified) {
+		entry->cursor_theme.color = entry->input_theme.foreground_color;
+	}
+	if (!entry->cursor_theme.text_color_specified) {
+		entry->cursor_theme.text_color = entry->background_color;
+	}
 
 	/*
 	 * Perform an initial render of the text.
