@@ -5,6 +5,7 @@
 #include <string.h>
 
 #include "fuzzy_match.h"
+#include "unicode.h"
 #include "xmalloc.h"
 
 #undef MAX
@@ -19,6 +20,7 @@ static int32_t fuzzy_match_recurse(
 		const char *restrict pattern,
 		const char *restrict str,
 		int32_t score,
+		bool first_match_only,
 		bool first_char);
 
 /*
@@ -30,10 +32,10 @@ int32_t fuzzy_match_simple_words(const char *restrict patterns, const char *rest
 {
 	int32_t score = 0;
 	char *saveptr = NULL;
-	char *tmp = xstrdup(patterns);
+	char *tmp = utf8_normalize(patterns);
 	char *pattern = strtok_r(tmp, " ", &saveptr);
 	while (pattern != NULL) {
-		char *c = strcasestr(str, pattern);
+		char *c = utf8_strcasestr(str, pattern);
 		if (c == NULL) {
 			score = INT32_MIN;
 			break;
@@ -55,7 +57,7 @@ int32_t fuzzy_match_words(const char *restrict patterns, const char *restrict st
 {
 	int32_t score = 0;
 	char *saveptr = NULL;
-	char *tmp = xstrdup(patterns);
+	char *tmp = utf8_normalize(patterns);
 	char *pattern = strtok_r(tmp, " ", &saveptr);
 	while (pattern != NULL) {
 		int32_t word_score = fuzzy_match(pattern, str);
@@ -78,8 +80,8 @@ int32_t fuzzy_match_words(const char *restrict patterns, const char *restrict st
 int32_t fuzzy_match(const char *restrict pattern, const char *restrict str)
 {
 	const int unmatched_letter_penalty = -1;
-	const size_t slen = strlen(str);
-	const size_t plen = strlen(pattern);
+	const size_t slen = utf8_strlen(str);
+	const size_t plen = utf8_strlen(pattern);
 	int32_t score = 0;
 
 	if (*pattern == '\0') {
@@ -92,8 +94,32 @@ int32_t fuzzy_match(const char *restrict pattern, const char *restrict str)
 	/* We can already penalise any unused letters. */
 	score += unmatched_letter_penalty * (int32_t)(slen - plen);
 
+        /*
+         * If the string is more than 100 characters, just find the first fuzzy
+         * match rather than the best.
+         *
+         * This is required as the number of possible matches (for patterns and
+         * strings all consisting of one letter) scales something like:
+         *
+         * slen! / (plen! (slen - plen)!)  ~  slen^plen for plen << slen
+         *
+         * This quickly grinds everything to a halt. 100 is chosen fairly
+         * arbitrarily from the following logic:
+         *
+         * - e is the most common character in English, at around 13% of
+         *   letters. Depending on the context, let's say this be up to 20%.
+         * - 100 * 0.20 = 20 repeats of the same character.
+         * - In the worst case here, 20! / (10! 10!) ~200,000 possible matches,
+         *   which is "slow but not frozen" for my machine.
+         *
+         * In reality, this worst case shouldn't be hit, and finding the "best"
+         * fuzzy match in lines of text > 100 characters isn't really in scope
+         * for a dmenu clone.
+         */
+        bool first_match_only = slen > 100;
+
 	/* Perform the match. */
-	score = fuzzy_match_recurse(pattern, str, score, true);
+	score = fuzzy_match_recurse(pattern, str, score, first_match_only, true);
 
 	return score;
 }
@@ -111,6 +137,7 @@ int32_t fuzzy_match_recurse(
 		const char *restrict pattern,
 		const char *restrict str,
 		int32_t score,
+		bool first_match_only,
 		bool first_char)
 {
 	if (*pattern == '\0') {
@@ -119,7 +146,7 @@ int32_t fuzzy_match_recurse(
 	}
 
 	const char *match = str;
-	const char search[2] = { *pattern, '\0' };
+	uint32_t search = utf8_to_utf32(pattern);
 
 	int32_t best_score = INT32_MIN;
 
@@ -127,14 +154,23 @@ int32_t fuzzy_match_recurse(
 	 * Find all occurrences of the next pattern character in str, and
 	 * recurse on them.
 	 */
-	while ((match = strcasestr(match, search)) != NULL) {
+	while ((match = utf8_strcasechr(match, search)) != NULL) {
+		int32_t jump = 0;
+		for (const char *tmp = str; tmp != match; tmp = utf8_next_char(tmp)) {
+			jump++;
+		}
 		int32_t subscore = fuzzy_match_recurse(
-				pattern + 1,
-				match + 1,
-				compute_score(match - str, first_char, match),
+				utf8_next_char(pattern),
+				utf8_next_char(match),
+				compute_score(jump, first_char, match),
+				first_match_only,
 				false);
 		best_score = MAX(best_score, subscore);
-		match++;
+		match = utf8_next_char(match);
+
+		if (first_match_only) {
+			break;
+		}
 	}
 
 	if (best_score == INT32_MIN) {
@@ -172,15 +208,18 @@ int32_t compute_score(int32_t jump, bool first_char, const char *restrict match)
 
 	int32_t score = 0;
 
+	const uint32_t cur = utf8_to_utf32(match);
+
 	/* Apply bonuses. */
 	if (!first_char && jump == 0) {
 		score += adjacency_bonus;
 	}
 	if (!first_char || jump > 0) {
-		if (isupper(*match) && islower(*(match - 1))) {
+		const uint32_t prev = utf8_to_utf32(utf8_prev_char(match));
+		if (utf32_isupper(cur) && utf32_islower(prev)) {
 			score += camel_bonus;
 		}
-		if (isalnum(*match) && !isalnum(*(match - 1))) {
+		if (utf32_isalnum(cur) && !utf32_isalnum(prev)) {
 			score += separator_bonus;
 		}
 	}
