@@ -16,6 +16,10 @@ static void clear_input(struct tofi *tofi);
 static void paste(struct tofi *tofi);
 static void select_previous_result(struct tofi *tofi);
 static void select_next_result(struct tofi *tofi);
+static void select_previous_page(struct tofi *tofi);
+static void select_next_page(struct tofi *tofi);
+static void next_cursor_or_result(struct tofi *tofi);
+static void previous_cursor_or_result(struct tofi *tofi);
 static void reset_selection(struct tofi *tofi);
 
 void input_handle_keypress(struct tofi *tofi, xkb_keycode_t keycode)
@@ -62,8 +66,12 @@ void input_handle_keypress(struct tofi *tofi, xkb_keycode_t keycode)
 		   )
 	{
 		paste(tofi);
+	} else if (sym == XKB_KEY_Left) {
+		previous_cursor_or_result(tofi);
+	} else if (sym == XKB_KEY_Right) {
+		next_cursor_or_result(tofi);
 	} else if (sym == XKB_KEY_Up || sym == XKB_KEY_Left || sym == XKB_KEY_ISO_Left_Tab
-			|| (key == KEY_K
+			|| ((key == KEY_K || key == KEY_P)
 				&& xkb_state_mod_name_is_active(
 					tofi->xkb_state,
 					XKB_MOD_NAME_CTRL,
@@ -72,7 +80,7 @@ void input_handle_keypress(struct tofi *tofi, xkb_keycode_t keycode)
 	   ) {
 		select_previous_result(tofi);
 	} else if (sym == XKB_KEY_Down || sym == XKB_KEY_Right || sym == XKB_KEY_Tab
-			|| (key == KEY_J
+			|| ((key == KEY_J || key == KEY_N)
 				&& xkb_state_mod_name_is_active(
 					tofi->xkb_state,
 					XKB_MOD_NAME_CTRL,
@@ -82,6 +90,10 @@ void input_handle_keypress(struct tofi *tofi, xkb_keycode_t keycode)
 		select_next_result(tofi);
 	} else if (sym == XKB_KEY_Home) {
 		reset_selection(tofi);
+	} else if (sym == XKB_KEY_Page_Up) {
+		select_previous_page(tofi);
+	} else if (sym == XKB_KEY_Page_Down) {
+		select_next_page(tofi);
 	} else if (sym == XKB_KEY_Escape
 			|| (key == KEY_C
 				&& xkb_state_mod_name_is_active(
@@ -98,6 +110,10 @@ void input_handle_keypress(struct tofi *tofi, xkb_keycode_t keycode)
 		return;
 	}
 
+	if (tofi->auto_accept_single && tofi->window.entry.results.count == 1) {
+		tofi->submit = true;
+	}
+
 	tofi->window.surface.redraw = true;
 }
 
@@ -122,24 +138,38 @@ void add_character(struct tofi *tofi, xkb_keycode_t keycode)
 			keycode,
 			buf,
 			sizeof(buf));
-	entry->input_utf32[entry->input_utf32_length] = utf8_to_utf32(buf);
-	entry->input_utf32_length++;
-	entry->input_utf32[entry->input_utf32_length] = U'\0';
-	memcpy(&entry->input_utf8[entry->input_utf8_length],
-			buf,
-			N_ELEM(buf));
-	entry->input_utf8_length += len;
-	if (entry->drun) {
-		struct string_ref_vec results = desktop_vec_filter(&entry->apps, entry->input_utf8, tofi->fuzzy_match);
-		string_ref_vec_destroy(&entry->results);
-		entry->results = results;
+	if (entry->cursor_position == entry->input_utf32_length) {
+		entry->input_utf32[entry->input_utf32_length] = utf8_to_utf32(buf);
+		entry->input_utf32_length++;
+		entry->input_utf32[entry->input_utf32_length] = U'\0';
+		memcpy(&entry->input_utf8[entry->input_utf8_length],
+				buf,
+				N_ELEM(buf));
+		entry->input_utf8_length += len;
+
+		if (entry->drun) {
+			struct string_ref_vec results = desktop_vec_filter(&entry->apps, entry->input_utf8, tofi->fuzzy_match);
+			string_ref_vec_destroy(&entry->results);
+			entry->results = results;
+		} else {
+			struct string_ref_vec tmp = entry->results;
+			entry->results = string_ref_vec_filter(&entry->results, entry->input_utf8, tofi->fuzzy_match);
+			string_ref_vec_destroy(&tmp);
+		}
+
+		reset_selection(tofi);
 	} else {
-		struct string_ref_vec tmp = entry->results;
-		entry->results = string_ref_vec_filter(&entry->results, entry->input_utf8, tofi->fuzzy_match);
-		string_ref_vec_destroy(&tmp);
+		for (size_t i = entry->input_utf32_length; i > entry->cursor_position; i--) {
+			entry->input_utf32[i] = entry->input_utf32[i - 1];
+		}
+		entry->input_utf32[entry->cursor_position] = utf8_to_utf32(buf);
+		entry->input_utf32_length++;
+		entry->input_utf32[entry->input_utf32_length] = U'\0';
+
+		input_refresh_results(tofi);
 	}
 
-	reset_selection(tofi);
+	entry->cursor_position++;
 }
 
 void input_refresh_results(struct tofi *tofi)
@@ -173,8 +203,20 @@ void delete_character(struct tofi *tofi)
 		return;
 	}
 
-	entry->input_utf32_length--;
-	entry->input_utf32[entry->input_utf32_length] = U'\0';
+	if (entry->cursor_position == 0) {
+		return;
+	} else if (entry->cursor_position == entry->input_utf32_length) {
+		entry->cursor_position--;
+		entry->input_utf32_length--;
+		entry->input_utf32[entry->input_utf32_length] = U'\0';
+	} else {
+		for (size_t i = entry->cursor_position - 1; i < entry->input_utf32_length - 1; i++) {
+			entry->input_utf32[i] = entry->input_utf32[i + 1];
+		}
+		entry->cursor_position--;
+		entry->input_utf32_length--;
+		entry->input_utf32[entry->input_utf32_length] = U'\0';
+	}
 
 	input_refresh_results(tofi);
 }
@@ -183,19 +225,26 @@ void delete_word(struct tofi *tofi)
 {
 	struct entry *entry = &tofi->window.entry;
 
-	if (entry->input_utf32_length == 0) {
+	if (entry->cursor_position == 0) {
 		/* No input to delete. */
 		return;
 	}
 
-	while (entry->input_utf32_length > 0 && utf32_isspace(entry->input_utf32[entry->input_utf32_length - 1])) {
-		entry->input_utf32_length--;
+	uint32_t new_cursor_pos = entry->cursor_position;
+	while (new_cursor_pos > 0 && utf32_isspace(entry->input_utf32[new_cursor_pos - 1])) {
+		new_cursor_pos--;
 	}
-	while (entry->input_utf32_length > 0 && !utf32_isspace(entry->input_utf32[entry->input_utf32_length - 1])) {
-		entry->input_utf32_length--;
+	while (new_cursor_pos > 0 && !utf32_isspace(entry->input_utf32[new_cursor_pos - 1])) {
+		new_cursor_pos--;
 	}
+	uint32_t new_length = entry->input_utf32_length - (entry->cursor_position - new_cursor_pos);
+	for (size_t i = 0; i < new_length; i++) {
+		entry->input_utf32[new_cursor_pos + i] = entry->input_utf32[entry->cursor_position + i];
+	}
+	entry->input_utf32_length = new_length;
 	entry->input_utf32[entry->input_utf32_length] = U'\0';
 
+	entry->cursor_position = new_cursor_pos;
 	input_refresh_results(tofi);
 }
 
@@ -203,6 +252,7 @@ void clear_input(struct tofi *tofi)
 {
 	struct entry *entry = &tofi->window.entry;
 
+	entry->cursor_position = 0;
 	entry->input_utf32_length = 0;
 	entry->input_utf32[0] = U'\0';
 
@@ -270,3 +320,53 @@ void select_next_result(struct tofi *tofi)
 		entry->last_num_results_drawn = entry->num_results_drawn;
 	}
 }
+
+void previous_cursor_or_result(struct tofi *tofi)
+{
+	struct entry *entry = &tofi->window.entry;
+
+	if (entry->cursor_theme.show
+			&& entry->selection == 0
+			&& entry->cursor_position > 0) {
+		entry->cursor_position--;
+	} else {
+		select_previous_result(tofi);
+	}
+}
+
+void next_cursor_or_result(struct tofi *tofi)
+{
+	struct entry *entry = &tofi->window.entry;
+
+	if (entry->cursor_theme.show
+			&& entry->cursor_position < entry->input_utf32_length) {
+		entry->cursor_position++;
+	} else {
+		select_next_result(tofi);
+	}
+}
+
+void select_previous_page(struct tofi *tofi)
+{
+	struct entry *entry = &tofi->window.entry;
+
+	if (entry->first_result >= entry->last_num_results_drawn) {
+		entry->first_result -= entry->last_num_results_drawn;
+	} else {
+		entry->first_result = 0;
+	}
+	entry->selection = 0;
+	entry->last_num_results_drawn = entry->num_results_drawn;
+}
+
+void select_next_page(struct tofi *tofi)
+{
+	struct entry *entry = &tofi->window.entry;
+
+	entry->first_result += entry->num_results_drawn;
+	if (entry->first_result >= entry->results.count) {
+		entry->first_result = 0;
+	}
+	entry->selection = 0;
+	entry->last_num_results_drawn = entry->num_results_drawn;
+}
