@@ -18,15 +18,16 @@
 #include "drun.h"
 #include "config.h"
 #include "entry.h"
-#include "image.h"
 #include "input.h"
 #include "log.h"
 #include "nelem.h"
 #include "lock.h"
+#include "scale.h"
 #include "shm.h"
 #include "string_vec.h"
 #include "string_vec.h"
 #include "unicode.h"
+#include "viewporter.h"
 #include "xmalloc.h"
 
 #undef MAX
@@ -97,18 +98,20 @@ static void zwlr_layer_surface_configure(
 		log_debug("Layer surface configure with no width or height.\n");
 		return;
 	}
-	log_debug("Layer surface configure, %d x %d.\n", width, height);
+	log_debug("Layer surface configure, %u x %u.\n", width, height);
 
 	/*
 	 * Resize the main window.
 	 * We want actual pixel width / height, so we have to scale the
 	 * values provided by Wayland.
 	 */
-	tofi->window.width = width * tofi->window.scale;
-	tofi->window.height = height * tofi->window.scale;
-
-	tofi->window.surface.width = tofi->window.width;
-	tofi->window.surface.height = tofi->window.height;
+	if (tofi->window.fractional_scale != 0) {
+		tofi->window.surface.width = scale_apply(width, tofi->window.fractional_scale);
+		tofi->window.surface.height = scale_apply(height, tofi->window.fractional_scale);
+	} else {
+		tofi->window.surface.width = width * tofi->window.scale;
+		tofi->window.surface.height = height * tofi->window.scale;
+	}
 
 	zwlr_layer_surface_v1_ack_configure(
 			tofi->window.zwlr_layer_surface,
@@ -697,6 +700,20 @@ static void registry_global(
 				&zwlr_layer_shell_v1_interface,
 				version);
 		log_debug("Bound to zwlr_layer_shell_v1 %u.\n", name);
+	} else if (!strcmp(interface, wp_viewporter_interface.name)) {
+		tofi->wp_viewporter = wl_registry_bind(
+				wl_registry,
+				name,
+				&wp_viewporter_interface,
+				1);
+		log_debug("Bound to wp_viewporter %u.\n", name);
+	} else if (!strcmp(interface, wp_fractional_scale_manager_v1_interface.name)) {
+		tofi->wp_fractional_scale_manager = wl_registry_bind(
+				wl_registry,
+				name,
+				&wp_fractional_scale_manager_v1_interface,
+				1);
+		log_debug("Bound to wp_fractional_scale_manager_v1 %u.\n", name);
 	}
 }
 
@@ -761,6 +778,19 @@ static const struct zwlr_layer_surface_v1_listener dummy_layer_surface_listener
 	.closed = dummy_layer_surface_close
 };
 
+static void dummy_fractional_scale_preferred_scale(
+		void *data,
+		struct wp_fractional_scale_v1 *wp_fractional_scale,
+		uint32_t scale)
+{
+	struct tofi *tofi = data;
+	tofi->window.fractional_scale = scale;
+}
+
+static const struct wp_fractional_scale_v1_listener dummy_fractional_scale_listener = {
+	.preferred_scale = dummy_fractional_scale_preferred_scale
+};
+
 static void dummy_surface_enter(
 		void *data,
 		struct wl_surface *wl_surface,
@@ -832,6 +862,12 @@ const struct option long_options[] = {
 	{"selection-background-corner-radius", required_argument, NULL, 0},
 	{"outline-width", required_argument, NULL, 0},
 	{"outline-color", required_argument, NULL, 0},
+	{"text-cursor", required_argument, NULL, 0},
+	{"text-cursor-style", required_argument, NULL, 0},
+	{"text-cursor-color", required_argument, NULL, 0},
+	{"text-cursor-background", required_argument, NULL, 0},
+	{"text-cursor-corner-radius", required_argument, NULL, 0},
+	{"text-cursor-thickness", required_argument, NULL, 0},
 	{"prompt-text", required_argument, NULL, 0},
 	{"prompt-padding", required_argument, NULL, 0},
 	{"prompt-color", required_argument, NULL, 0},
@@ -877,6 +913,7 @@ const struct option long_options[] = {
 	{"history-file", required_argument, NULL, 0},
 	{"fuzzy-match", required_argument, NULL, 0},
 	{"require-match", required_argument, NULL, 0},
+	{"auto-accept-single", required_argument, NULL, 0},
 	{"hide-input", required_argument, NULL, 0},
 	{"hidden-character", required_argument, NULL, 0},
 	{"drun-launch", required_argument, NULL, 0},
@@ -935,14 +972,18 @@ static void parse_args(struct tofi *tofi, int argc, char *argv[])
 	opt = getopt_long(argc, argv, short_options, long_options, &option_index);
 	while (opt != -1) {
 		if (opt == 0) {
-			config_apply(tofi, long_options[option_index].name, optarg);
+			if (!config_apply(tofi, long_options[option_index].name, optarg)) {
+				exit(EXIT_FAILURE);
+			}
 		} else if (opt == 'k') {
 			/*
 			 * Backwards compatibility for --late-keyboard-init not
 			 * taking an argument.
 			 */
 			if (optarg) {
-				config_apply(tofi, long_options[option_index].name, optarg);
+				if (!config_apply(tofi, long_options[option_index].name, optarg)) {
+					exit(EXIT_FAILURE);
+				}
 			} else {
 				tofi->late_keyboard_init = true;
 			}
@@ -1015,12 +1056,22 @@ static bool do_submit(struct tofi *tofi)
 static void read_clipboard(struct tofi *tofi)
 {
 	struct entry *entry = &tofi->window.entry;
+
+	/* Make a copy of any text after the cursor. */
+	uint32_t *end_text = NULL;
+	size_t end_text_length = entry->input_utf32_length - entry->cursor_position;
+	if (end_text_length > 0) {
+		end_text = xcalloc(end_text_length, sizeof(*entry->input_utf32));
+		memcpy(end_text,
+				&entry->input_utf32[entry->cursor_position],
+				end_text_length * sizeof(*entry->input_utf32));
+	}
 	/* Buffer for 4 UTF-8 bytes plus a null terminator. */
 	char buffer[5];
 	memset(buffer, 0, N_ELEM(buffer));
 	errno = 0;
 	bool eof = false;
-	while (entry->input_utf32_length < N_ELEM(entry->input_utf32)) {
+	while (entry->cursor_position < N_ELEM(entry->input_utf32)) {
 		for (size_t i = 0; i < 4; i++) {
 			/*
 			 * Read input 1 byte at a time. This is slow, but easy,
@@ -1058,8 +1109,8 @@ static void read_clipboard(struct tofi *tofi)
 				log_error("Invalid UTF-8 character in clipboard: %s\n", buffer);
 				break;
 			} else {
-				entry->input_utf32[entry->input_utf32_length] = unichar;
-				entry->input_utf32_length++;
+				entry->input_utf32[entry->cursor_position] = unichar;
+				entry->cursor_position++;
 				break;
 			}
 		}
@@ -1068,6 +1119,19 @@ static void read_clipboard(struct tofi *tofi)
 			break;
 		}
 	}
+	entry->input_utf32_length = entry->cursor_position;
+
+	/* If there was any text after the cursor, re-insert it now. */
+	if (end_text != NULL) {
+		for (size_t i = 0; i < end_text_length; i++) {
+			if (entry->input_utf32_length == N_ELEM(entry->input_utf32)) {
+				break;
+			}
+			entry->input_utf32[entry->input_utf32_length] = end_text[i];
+			entry->input_utf32_length++;
+		}
+		free(end_text);
+	}
 	entry->input_utf32[MIN(entry->input_utf32_length, N_ELEM(entry->input_utf32) - 1)] = U'\0';
 
 	clipboard_finish_paste(&tofi->clipboard);
@@ -1113,7 +1177,8 @@ int main(int argc, char *argv[])
 				.placeholder_theme.foreground_color = {1.0f, 1.0f, 1.0f, 0.66f},
 				.placeholder_theme.foreground_specified = true,
 				.selection_theme.foreground_color = {0.976f, 0.149f, 0.447f, 1.0f},
-				.selection_theme.foreground_specified = true
+				.selection_theme.foreground_specified = true,
+				.cursor_theme.thickness = 2
 			}
 		},
 		.anchor =  ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP
@@ -1134,7 +1199,7 @@ int main(int argc, char *argv[])
 	}
 
 	parse_args(&tofi, argc, argv);
-	log_debug("Config done\n");
+	log_debug("Config done.\n");
 
 	if (!tofi.multiple_instance && lock_check()) {
 		log_error("Another instance of tofi is already running.\n");
@@ -1190,20 +1255,28 @@ int main(int argc, char *argv[])
 	log_unindent();
 	log_debug("Second roundtrip done.\n");
 
-	/* If there's more than one output, we need to select one. */
-	if (wl_list_length(&tofi.output_list) > 1) {
+	{
 		/*
-		 * Determine the default output
+		 * Determine the output we're going to appear on, and get its
+		 * fractional scale if supported.
 		 *
 		 * This seems like an ugly solution, but as far as I know
 		 * there's no way to determine the default output other than to
 		 * call get_layer_surface with NULL as the output and see which
 		 * output our surface turns up on.
 		 *
+		 * Additionally, determining fractional scale factors can
+		 * currently only be done by attaching a wp_fractional_scale to
+		 * a surface and displaying it.
+		 *
 		 * Here we set up a single pixel surface, perform the required
 		 * two roundtrips, then tear it down. tofi.default_output
-		 * should then contain the output our surface was assigned to.
+		 * should then contain the output our surface was assigned to,
+		 * and tofi.window.fractional_scale should have the scale
+		 * factor.
 		 */
+		log_debug("Determining output.\n");
+		log_indent();
 		struct surface surface = {
 			.width = 1,
 			.height = 1
@@ -1215,11 +1288,38 @@ int main(int argc, char *argv[])
 				&dummy_surface_listener,
 				&tofi);
 
+		struct wp_fractional_scale_v1 *wp_fractional_scale = NULL;
+		if (tofi.wp_fractional_scale_manager != NULL) {
+			wp_fractional_scale =
+				wp_fractional_scale_manager_v1_get_fractional_scale(
+						tofi.wp_fractional_scale_manager,
+						surface.wl_surface);
+			wp_fractional_scale_v1_add_listener(
+					wp_fractional_scale,
+					&dummy_fractional_scale_listener,
+					&tofi);
+		}
+
+		/*
+		 * If we have a desired output, make sure we appear on it so we
+		 * can determine the correct fractional scale.
+		 */
+		struct wl_output *wl_output = NULL;
+		if (tofi.target_output_name[0] != '\0') {
+			struct output_list_element *el;
+			wl_list_for_each(el, &tofi.output_list, link) {
+				if (!strcmp(tofi.target_output_name, el->name)) {
+					wl_output = el->wl_output;
+					break;
+				}
+			}
+		}
+
 		struct zwlr_layer_surface_v1 *zwlr_layer_surface =
 			zwlr_layer_shell_v1_get_layer_surface(
 					tofi.zwlr_layer_shell,
 					surface.wl_surface,
-					NULL,
+					wl_output,
 					ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND,
 					"dummy");
 		/*
@@ -1239,12 +1339,27 @@ int main(int argc, char *argv[])
 				1,
 				1);
 		wl_surface_commit(surface.wl_surface);
+		log_debug("First dummy roundtrip start.\n");
+		log_indent();
 		wl_display_roundtrip(tofi.wl_display);
+		log_unindent();
+		log_debug("First dummy roundtrip done.\n");
+		log_debug("Initialising dummy surface.\n");
+		log_indent();
 		surface_init(&surface, tofi.wl_shm);
 		surface_draw(&surface);
+		log_unindent();
+		log_debug("Dummy surface initialised.\n");
+		log_debug("Second dummy roundtrip start.\n");
+		log_indent();
 		wl_display_roundtrip(tofi.wl_display);
+		log_unindent();
+		log_debug("Second dummy roundtrip done.\n");
 		surface_destroy(&surface);
 		zwlr_layer_surface_v1_destroy(zwlr_layer_surface);
+		if (wp_fractional_scale != NULL) {
+			wp_fractional_scale_v1_destroy(wp_fractional_scale);
+		}
 		wl_surface_destroy(surface.wl_surface);
 
 		/*
@@ -1286,13 +1401,11 @@ int main(int argc, char *argv[])
 				free(el);
 			}
 		}
-	}
-	{
+
 		/*
 		 * The only output left should either be the one we want, or
 		 * the first that was advertised.
 		 */
-		struct output_list_element *el;
 		el = wl_container_of(tofi.output_list.next, el, link);
 
 		/*
@@ -1313,6 +1426,7 @@ int main(int argc, char *argv[])
 		}
 		tofi.window.scale = el->scale;
 		tofi.window.transform = el->transform;
+		log_unindent();
 		log_debug("Selected output %s.\n", el->name);
 	}
 
@@ -1385,6 +1499,12 @@ int main(int argc, char *argv[])
 	}
 	tofi.window.entry.results = string_ref_vec_copy(&tofi.window.entry.commands);
 
+	if (tofi.auto_accept_single && tofi.window.entry.results.count == 1) {
+		log_debug("Only one result, exiting.\n");
+		do_submit(&tofi);
+		return EXIT_SUCCESS;
+	}
+
 	/*
 	 * Next, we create the Wayland surface, which takes on the
 	 * layer shell role.
@@ -1396,9 +1516,43 @@ int main(int argc, char *argv[])
 			tofi.window.surface.wl_surface,
 			&wl_surface_listener,
 			&tofi);
-	wl_surface_set_buffer_scale(
-			tofi.window.surface.wl_surface,
-			tofi.window.scale);
+	if (tofi.window.width == 0 || tofi.window.height == 0) {
+		/*
+		 * Workaround for compatibility with legacy behaviour.
+		 *
+		 * Before the fractional_scale protocol was released, there was
+		 * no way for a client to know whether a fractional scale
+		 * factor had been set, meaning percentage-based dimensions
+		 * were incorrect. As a workaround for full-size windows, we
+		 * allowed specifying 0 for the width / height, which caused
+		 * zwlr_layer_shell to tell us the correct size to use.
+		 *
+		 * To make fractional scaling work, we have to use
+		 * wp_viewporter, and no longer need to set the buffer scale.
+		 * However, viewporter doesn't allow specifying 0 for
+		 * destination width or height. As a workaround, if 0 size is
+		 * set, don't use viewporter, warn the user and set the buffer
+		 * scale here.
+		 */
+		log_warning("Width or height set to 0, disabling fractional scaling support.\n");
+		log_warning("If your compositor supports the fractional scale protocol, percentages are preferred.\n");
+		tofi.window.fractional_scale = 0;
+		wl_surface_set_buffer_scale(
+				tofi.window.surface.wl_surface,
+				tofi.window.scale);
+	} else if (tofi.wp_viewporter == NULL) {
+		/*
+		 * We also could be running on a Wayland compositor which
+		 * doesn't support wp_viewporter, in which case we need to use
+		 * the old scaling method.
+		 */
+		log_warning("Using an outdated compositor, "
+				"fractional scaling will not work properly.\n");
+		tofi.window.fractional_scale = 0;
+		wl_surface_set_buffer_scale(
+				tofi.window.surface.wl_surface,
+				tofi.window.scale);
+	}
 
 	/* Grab the first (and only remaining) output from our list. */
 	struct wl_output *wl_output;
@@ -1416,8 +1570,7 @@ int main(int argc, char *argv[])
 			"launcher");
 	zwlr_layer_surface_v1_set_keyboard_interactivity(
 			tofi.window.zwlr_layer_surface,
-			ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE
-			);
+			ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE);
 	zwlr_layer_surface_v1_add_listener(
 			tofi.window.zwlr_layer_surface,
 			&zwlr_layer_surface_listener,
@@ -1428,16 +1581,38 @@ int main(int argc, char *argv[])
 	zwlr_layer_surface_v1_set_exclusive_zone(
 			tofi.window.zwlr_layer_surface,
 			tofi.window.exclusive_zone);
-	zwlr_layer_surface_v1_set_size(
-			tofi.window.zwlr_layer_surface,
-			tofi.window.width / tofi.window.scale,
-			tofi.window.height / tofi.window.scale);
 	zwlr_layer_surface_v1_set_margin(
 			tofi.window.zwlr_layer_surface,
-			tofi.window.margin_top / tofi.window.scale,
-			tofi.window.margin_right / tofi.window.scale,
-			tofi.window.margin_bottom / tofi.window.scale,
-			tofi.window.margin_left / tofi.window.scale);
+			tofi.window.margin_top,
+			tofi.window.margin_right,
+			tofi.window.margin_bottom,
+			tofi.window.margin_left);
+	/*
+	 * No matter whether we're scaling via Cairo or not, we're presenting a
+	 * scaled buffer to Wayland, so scale the window size here if we
+	 * haven't already done so.
+	 */
+	zwlr_layer_surface_v1_set_size(
+			tofi.window.zwlr_layer_surface,
+			tofi.window.width,
+			tofi.window.height);
+
+	/*
+	 * Set up a viewport for our surface, necessary for fractional scaling.
+	 */
+	if (tofi.wp_viewporter != NULL) {
+		tofi.window.wp_viewport = wp_viewporter_get_viewport(
+				tofi.wp_viewporter,
+				tofi.window.surface.wl_surface);
+		if (tofi.window.width > 0 && tofi.window.height > 0) {
+			wp_viewport_set_destination(
+					tofi.window.wp_viewport,
+					tofi.window.width,
+					tofi.window.height);
+		}
+	}
+
+	/* Commit the surface to finalise setup. */
 	wl_surface_commit(tofi.window.surface.wl_surface);
 
 	/*
@@ -1452,7 +1627,6 @@ int main(int argc, char *argv[])
 			&wl_data_device_listener,
 			&tofi.clipboard);
 
-
 	/*
 	 * Now that we've done all our Wayland-related setup, we do another
 	 * roundtrip. This should cause the layer surface window to be
@@ -1487,11 +1661,28 @@ int main(int argc, char *argv[])
 	 */
 	log_debug("Initialising renderer.\n");
 	log_indent();
-	entry_init(
-			&tofi.window.entry,
-			tofi.window.surface.shm_pool_data,
-			tofi.window.width,
-			tofi.window.height);
+	{
+		/*
+		 * No matter how we're scaling (with fractions, integers or not
+		 * at all), we pass a fractional scale factor (the numerator of
+		 * a fraction with denominator 120) to our setup function for
+		 * ease.
+		 */
+		uint32_t scale = 120;
+		if (tofi.use_scale) {
+			if (tofi.window.fractional_scale != 0) {
+				scale = tofi.window.fractional_scale;
+			} else {
+				scale = tofi.window.scale * 120;
+			}
+		}
+		entry_init(
+				&tofi.window.entry,
+				tofi.window.surface.shm_pool_data,
+				tofi.window.surface.width,
+				tofi.window.surface.height,
+				scale);
+	}
 	log_unindent();
 	log_debug("Renderer initialised.\n");
 
@@ -1507,11 +1698,13 @@ int main(int argc, char *argv[])
 	 * possible after startup.
 	 */
 	wl_display_roundtrip(tofi.wl_display);
+	log_debug("Initialising second buffer.\n");
 	memcpy(
 		cairo_image_surface_get_data(tofi.window.entry.cairo[1].surface),
 		cairo_image_surface_get_data(tofi.window.entry.cairo[0].surface),
-		tofi.window.entry.image.width * tofi.window.entry.image.height * sizeof(uint32_t)
+		tofi.window.surface.width * tofi.window.surface.height * sizeof(uint32_t)
 	);
+	log_debug("Second buffer initialised.\n");
 
 	/* We've just rendered, so we don't need to do it again right now. */
 	tofi.window.surface.redraw = false;
@@ -1658,6 +1851,9 @@ int main(int argc, char *argv[])
 	 */
 	surface_destroy(&tofi.window.surface);
 	entry_destroy(&tofi.window.entry);
+	if (tofi.window.wp_viewport != NULL) {
+		wp_viewport_destroy(tofi.window.wp_viewport);
+	}
 	zwlr_layer_surface_v1_destroy(tofi.window.zwlr_layer_surface);
 	wl_surface_destroy(tofi.window.surface.wl_surface);
 	if (tofi.wl_keyboard != NULL) {
@@ -1684,6 +1880,12 @@ int main(int argc, char *argv[])
 		}
 	}
 	wl_shm_destroy(tofi.wl_shm);
+	if (tofi.wp_fractional_scale_manager != NULL) {
+		wp_fractional_scale_manager_v1_destroy(tofi.wp_fractional_scale_manager);
+	}
+	if (tofi.wp_viewporter != NULL) {
+		wp_viewporter_destroy(tofi.wp_viewporter);
+	}
 	zwlr_layer_shell_v1_destroy(tofi.zwlr_layer_shell);
 	xkb_state_unref(tofi.xkb_state);
 	xkb_keymap_unref(tofi.xkb_keymap);
