@@ -8,9 +8,9 @@
 #include "clipboard.h"
 #include "color.h"
 #include "entry.h"
-#include "image.h"
 #include "surface.h"
 #include "wlr-layer-shell-unstable-v1.h"
+#include "fractional-scale-v1.h"
 
 #define MAX_OUTPUT_NAME_LEN 256
 #define MAX_TERMINAL_NAME_LEN 256
@@ -35,6 +35,8 @@ struct tofi {
 	struct wl_shm *wl_shm;
 	struct wl_data_device_manager *wl_data_device_manager;
 	struct wl_data_device *wl_data_device;
+	struct wp_viewporter *wp_viewporter;
+	struct wp_fractional_scale_manager_v1 *wp_fractional_scale_manager;
 	struct zwlr_layer_shell_v1 *zwlr_layer_shell;
 	struct wl_list output_list;
 	struct output_list_element *default_output;
@@ -57,11 +59,13 @@ struct tofi {
 	struct clipboard clipboard;
 	struct {
 		struct surface surface;
+		struct wp_viewport *wp_viewport;
 		struct zwlr_layer_surface_v1 *zwlr_layer_surface;
 		struct entry entry;
 		uint32_t width;
 		uint32_t height;
 		uint32_t scale;
+		uint32_t fractional_scale;
 		int32_t transform;
 		int32_t exclusive_zone;
 		int32_t margin_top;
@@ -95,6 +99,7 @@ struct tofi {
 	bool drun_print_exec;
 	bool fuzzy_match;
 	bool require_match;
+	bool auto_accept_single;
 	bool multiple_instance;
 	char target_output_name[MAX_OUTPUT_NAME_LEN];
 	char default_terminal[MAX_TERMINAL_NAME_LEN];
